"""
Kernel regression model class

------------
@attributes:
------------

    This-class attributes
    - X             : (numpy.ndarray) training data (feature matrix)
    - y             : (numpy.ndarray) training labels (response variable)
    - m             : (int) number of training examples (rows of X)
    - n             : (int) number of training features (cols of X)
    - gamma         : (float) kernel bandwidth parameter
    - gammas        : (numpy.ndarray/list) kernel bandwidths parameter set
    - kernel        : (string) kernel function specification
    - normalize     : (bool) data normalization specification
    - debug_model   : (bool) model debugging option

---------
@methods:
---------

    This-class methods
    - fit           : fit the kernel bandwidth of the model (minimize MSE)
    - predict       : predict the test data labels (response variable)

    - __init__      : initialize a class object (class constructor)
    - __repr__      : operator overloading (basic representation)
    - __str__       : operator overloading (string representation)

------------
@properties:
------------

    This-class properties
    - parameters    : get the model parameters string
    - display       : get the model properties display string

@author: Ing. Zoltan Galaz
@email:  z.galaz@phd.feec.vutbr.cz
@www: http://bdalab.utko.feec.vutbr.cz/

Last revision: 28.5.2017
"""

__version__ = '1.0'


import numpy as np
import matplotlib.pyplot as plt
from utilities import kernel_functions as kernels


# Default parameters
#
# GAMMA     : kernel bandwidth parameter used to compute data distances
# GAMASET   : predefined set of gamma values to optimize the model over
# KERNEL    : kernel function specification
# NORMALIZE : data normalization specification
# DEBUGMODE : debugging mode (graphs/print statements/etc.)
GAMMA     = 1.0
GAMMASET  = np.logspace(-5, 0.9, 20)
KERNEL    = 'radial_kernel'
NORMALIZE = True
DEBUGMODE = False


class KernelRegression(object):
    """
    Kernel regression model class

    ------------
    @attributes:
    ------------

        This-class attributes
        - X             : (numpy.ndarray) training data (feature matrix)
        - y             : (numpy.ndarray) training labels (response variable)
        - m             : (int) number of training examples (rows of X)
        - n             : (int) number of training features (cols of X)
        - gamma         : (float) kernel bandwidth parameter
        - gammas        : (numpy.ndarray/list) kernel bandwidths parameter set
        - kernel        : (string) kernel function specification
        - normalize     : (bool) data normalization specification
        - debug_model   : (bool) model debugging option

    ---------
    @methods:
    ---------

        This-class methods
        - fit           : fit the kernel bandwidth of the model (minimize MSE)
        - predict       : predict the test data labels (response variable)

        - __init__      : initialize a class object (class constructor)
        - __repr__      : operator overloading (basic representation)
        - __str__       : operator overloading (string representation)

    ------------
    @properties:
    ------------

        This-class properties
        - parameters    : get the model parameters string
        - display       : get the model properties display string
    """

    def __init__(self, X, y, **kwargs):
        """
        Initialize the Kernel Model object

        Construct the Kernel Model object using mandatory positional parameters that
        specify input training data matrix and the associated response variable. The
        function also reads additional keyword arguments that provide more control
        of the learning/evaluation process.

        The feature matrix is first converted to numpy.ndarray (2 dimensional). Next
        the vector of ones is appended (1. column) to specify the data for intercept
        term theta_0. The response variable is also converted to numpy.ndarray. The
        feature matrix (X) shape: rows = observations, cols = features.

        Args:
            X: (numpy.ndarray/list) training data feature matrix
            y: (numpy.ndarray/list) training data response variable (labels)

            **kwargs
            gamma: (float) kernel bandwidth parameter
            gammas: (numpy.ndarray/list) kernel bandwidths parameter set
            kernel: (string) kernel function specification
            normalize: (bool) data normalization specification
            debug_model: (bool) debugging mode specification

        Returns:
            self: (KernelRegression object) KernelRegression object
        """

        # Initialize optional parameters
        self.gamma = kwargs.get('gamma', GAMMA)
        self.gammas = kwargs.get('gammas', GAMMASET)
        self.kernel = kwargs.get('kernel', KERNEL)
        self.normalize = kwargs.get('normalize', NORMALIZE)
        self.debug_model = kwargs.get('debug_model', DEBUGMODE)

        # Pre-process mandatory parameters
        _X = np.array(X, ndmin=2)
        _y = np.array(y, ndmin=2)

        if _X.shape[0] != y.shape[0]:
            raise ValueError("X, y: number of observations does no match")
        if _y.shape[1] > 1:
            raise ValueError("y: number of dimensions exceeded")

        # Normalize the feature matrix
        if self.normalize:
            _X = (_X - np.mean(_X, axis=0)) / np.std(_X, axis=0)

        # Initialize mandatory parameters
        self.X = _X
        self.y = _y

        # Initialize the properties
        self.m = self.X.shape[0]
        self.n = self.X.shape[1]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.display

    # ----------------------------------------------------------------------------------------
    # Public functions definition
    # ----------------------------------------------------------------------------------------

    def fit(self):
        """
        Optimize the gamma parameter

        Fit the gamma parameter of the model. The gamma (bandwidth of the kernel) is
        optimized to minimize the mean-squared error (MSE) of the predicted response
        variable using the training data. The final gamma is set as the preferred
        parameter used during the prediction.

        Args:
            self: (KernelRegression object) KernelRegression object

        Returns:
            None
        """

        # Prepare the grid search
        _error = np.inf
        _gamma = self.gamma

        # Search over predefined gamma values
        for grid in self.gammas:
            self.gamma = grid

            # Compute the mean squared error
            _mse = ((self.y - self.predict()) ** 2).mean(axis=0)
            if _mse < _error:
                _error = _mse
                _gamma = self.gamma

            # Debug the learning process
            if self.debug_model:
                print("Fitting {:} "
                      "(gamma) : {:.4f}, "
                      "mse : {:}".format(self.kernel, grid, _mse))

        # Set the final gamma
        self.gamma = _gamma

        # Debug the learning process
        if self.debug_model:
            print(self.display)

    def predict(self, X_test=None):
        """
        Predict the response variable using test data

        Predict the response variable  for given test examples. The Response variables
        are predicted using the kernel-based distances between the training examples
        and the query data points. This function uses Euclidean distance metric and
        radial basis kernel defined as follows

        distance(x, y)   = $$ sqrt((x[0]-y[0])^2 + ... + (x[n-1]-y[n-1])^2) $$
        kernel(d, gamma) = $$ exp(-gamma * d^2) $$ (d = Euclidean distance)

        Euclidean distance <https://en.wikipedia.org/wiki/Euclidean_distance>
        RBF kernel <https://en.wikipedia.org/wiki/Radial_basis_function_kernel>

        Args:
            self: (KernelRegression object) KernelRegression object
            X_test: (numpy.ndarray/list) testing data feature matrix

        Returns:
            y_hat: (numpy.ndarray/list) predicted response variable
        """

        # Pre-process the testing data
        if X_test is not None:
            _X = np.array(X_test, ndmin=2)
            _m = np.mean(_X, axis=0)
            _s = np.std(_X, axis=0)

            # Normalize the data
            if self.normalize:
                _X = (_X - _m) / _s
        else:
            _X = self.X

        # Compute the pairwise kernel-based distances
        y_hat = np.zeros((_X.shape[0], 1))
        for i in range(0, _X.shape[0]):
            _query = _X[i,:]
            _label = self.y[i]

            # Compute the kernel-based distance
            if self.kernel == 'radial_kernel':
                _kernel = kernels.radial_kernel(self.X, _query, 1*self.gamma)
            if self.kernel == 'gaussian_kernel':
                _kernel = kernels.radial_kernel(self.X, _query, 1/self.gamma)
            if self.kernel == 'linear_kernel':
                _kernel = kernels.linear_kernel(self.X, _query)

            # Predict the response variable
            y_hat[i] = (_label * _kernel).sum() / _kernel.sum()

        # Return the predicted response variables
        return y_hat

    # ----------------------------------------------------------------------------------------
    # Properties definition
    # ----------------------------------------------------------------------------------------

    @property
    def parameters(self):
        """
        Get the model parameters string

        Get the model parameters string. The property return the following parameters
        X.shape (rows, cols): number of observations, number of features
        y.shape (rows, cols): number of observations, response dimension
        kernel: kernel function specification
        gamma: kernel bandwidth parameter
        normalization: data normalization specification

        Args:
            self: (KernelRegression object) KernelRegression object

        Returns:
            param: (String) model parameters
        """

        # Return the model parameters
        return ("X: shape      = ({:d}, {:d})\n".format(self.m, self.n) +
                "y: shape      = ({:d}, {:d})\n".format(self.m, 1) +
                "kernel        = {:}\n".format(self.kernel) +
                "kernel gamma  = {:.4f}\n".format(self.gamma) +
                "normalization = {:}\n".format(self.normalize))

    @property
    def display(self):
        """
        Get the model specification string

        Get the model specification string. the string contains information about
        model parameters (see: parameters) It is used by the __str()__ method to
        provide human-readable specification of the model.

        Args:
            self: (KernelRegression object) KernelRegression object

        Returns:
            disp: (String) model specification
        """

        # Return the model specification
        return ("\nModel parameters: \n" + self.parameters)


if __name__ == '__main__':
    import pandas as pd

    # Load the training data
    train_data = pd.read_csv(
        "data/train_data.csv",
        dtype={'bathrooms': float, 'waterfront': int, 'sqft_above': int,
               'sqft_living15': float, 'grade': int, 'yr_renovated': int,
               'price': float, 'bedrooms': float, 'zipcode': str, 'long': float,
               'sqft_lot15': float, 'sqft_living': float, 'floors': str,
               'condition': int, 'lat': float, 'date': str, 'sqft_basement': int,
               'yr_built': int, 'id': str, 'sqft_lot': int, 'view': int}
    )

    _specification = ['bathrooms', 'sqft_living15', 'bedrooms', 'sqft_lot15']
    _p1 = np.array(train_data[_specification[0]], ndmin=2)
    _p2 = np.array(train_data[_specification[1]], ndmin=2)
    _p3 = np.array(train_data[_specification[2]], ndmin=2)
    _p4 = np.array(train_data[_specification[3]], ndmin=2)
    X_train = np.transpose(np.concatenate((_p1, _p2, _p3, _p4), axis=0))
    y_train = np.transpose(np.array(train_data['price'], ndmin=2))

    # Load the testing data
    test_data = pd.read_csv(
        "data/test_data.csv",
        dtype={'bathrooms': float, 'waterfront': int, 'sqft_above': int,
               'sqft_living15': float, 'grade': int, 'yr_renovated': int,
               'price': float, 'bedrooms': float, 'zipcode': str, 'long': float,
               'sqft_lot15': float, 'sqft_living': float, 'floors': str,
               'condition': int, 'lat': float, 'date': str, 'sqft_basement': int,
               'yr_built': int, 'id': str, 'sqft_lot': int, 'view': int}
    )

    _specification = ['bathrooms', 'sqft_living15', 'bedrooms', 'sqft_lot15']
    _p1 = np.array(test_data[_specification[0]], ndmin=2)
    _p2 = np.array(test_data[_specification[1]], ndmin=2)
    _p3 = np.array(test_data[_specification[2]], ndmin=2)
    _p4 = np.array(test_data[_specification[3]], ndmin=2)
    X_test = np.transpose(np.concatenate((_p1, _p2, _p3, _p4), axis=0))
    y_test = np.transpose(np.array(test_data['price'], ndmin=2))

    # Build the kernel regression model
    ker_model = KernelRegression(
        X_train,
        y_train,
        normalize=True,
        debug_model=True)

    # Fit the regression model
    ker_model.fit()

    # Predict the testing data
    y_hat = ker_model.predict(X_test)

    # Plot the fitted data
    if ker_model.n <= 4:
        for i in range(0, ker_model.n):
            plt.subplot(2, 2, i+1)

            # Plot true labels
            h_true, = plt.plot(
                X_test[:, i],
                y_test[:, 0],
                'rx',
                markersize=10,
                label="y_test"
            )

            # Plot predicted labels
            h_hat, = plt.plot(
                X_test[:, i],
                y_hat[:, 0],
                'bx',
                markersize=10,
                label="y_hat"
            )

            # Set the properties
            plt.grid(True)
            plt.title('y = f(x[:,{}])'.format(str(i)))
            plt.ylabel('y')
            plt.xlabel('x')
            plt.legend(handles=[h_true, h_hat])

        plt.show()